from json import JSONEncoder
from pathlib import PosixPath

from .components import Connection, Parameter, ParameterGroup, ParameterType, SharedConfig
from .engine import Action, Engine


class ModelEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, PosixPath):
            return str(o)
        if isinstance(o, Connection):
            return self.serialize_connection(o)
        if isinstance(o, SharedConfig):
            return self.serialize_shared_config(o)
        if isinstance(o, Parameter):
            return self.serialize_parameter(o)
        if isinstance(o, ParameterGroup):
            return self.serialize_parameter_group(o)
        elif isinstance(o, ParameterType):
            return self.serialize_parameter_type(o)
        elif isinstance(o, Engine):
            return self.serialize_engine(o)
        elif isinstance(o, Action):
            return self.serialize_action(o)
        else:
            return super().default(o)

    def serialize_shared_config(self, shared_config: SharedConfig):
        result = {
            "uuid": shared_config.shared_config_type_uuid,
            "alias": shared_config.alias,
            "nnd_model_version": 2,
            "description": shared_config.description,
            "categories": [{"name": val} for val in shared_config.categories]
            if shared_config.categories
            else [{"name": "General"}],
            "type": "shared_config_type",
        }
        result["parameters"] = shared_config.parameter_groups
        return result

    def serialize_connection(self, connection: Connection):
        result = {
            "uuid": connection.connection_type_uuid,
            "alias": connection.alias,
            "nnd_model_version": 2,
            "description": connection.description,
            "categories": [{"name": val} for val in connection.categories]
            if connection.categories
            else [{"name": "General"}],
            "type": "connection",
        }
        result["parameters"] = connection.parameter_groups
        return result

    def serialize_engine(self, engine: Engine):
        parameter_categories = set()
        for action in engine.actions.values():
            for pg in action.parameter_groups.values():
                for parameter in [p for p in pg.parameters if p.categories]:
                    parameter_categories.update(
                        [cat["name"] for cat in parameter.categories]
                    )

        result = {
            "uuid": engine.uuid,
            "alias": engine.alias,
            "description": engine.description,
            "nnd_model_version": engine.nnd_model_version,
            "categories": engine.categories,
            "type": engine.model_type.__str__(),
            "parameter_categories": [{"name": c} for c in sorted(parameter_categories)],
        }
        if engine.icons:
            result["icons"] = engine.icons
        if engine.help:
            result["help"] = engine.help
        result["actions"] = engine.actions
        return result

    def serialize_action(self, action: Action):
        # add parameters to this dict last so
        # that it's more human readable
        result = {
            "display_name": action.display_name or action.name.capitalize(),
            "description": action.description,
        }
        if action.help:
            result["help"] = action.help
        result["parameters"] = [p for p in action.parameter_groups.values()]
        return result

    def serialize_parameter_group(self, pg: ParameterGroup):
        result = {
            "name": pg.name,
            "display_name": pg.display_name,
            "description": pg.description,
            "type": pg.type,
            "collapsed": pg.collapsed,
            "parameters": [p for p in pg.parameters],
        }
        if pg.shared_parameter_group_uuid:
            result["shared_parameter_group_uuid"] = pg.shared_parameter_group_uuid
        return result

    def serialize_parameter(self, p: Parameter):
        result = {
            "name": p.name,
            "display_name": p.display_name,
            "description": p.description,
            "required": p.required,
        }
        if p.help:
            result["help"] = p.help
        if p.default:
            result["default"] = p.type.dump(p.default)
        if p.many:
            result["many"] = p.many
        if p.categories:
            result["categories"] = p.categories
        if p.show_on:
            result["show_on"] = {
                "parameter_name": p.show_on.name,
                "parameter_value": p.show_on.value,
            }
        if isinstance(p.type, Connection):
            result.update(self.serialize_connection_parameter(p.type))
        elif isinstance(p.type, SharedConfig):
            result.update(self.serialize_shared_config_parameter(p.type))
        else:
            result.update(self.serialize_parameter_type(p.type))
        return result

    def serialize_connection_parameter(self, c: Connection):
        return {"connection_type_uuid": c.connection_type_uuid, "type": c.type}

    def serialize_shared_config_parameter(self, c: SharedConfig):
        return {"shared_config_type_uuid": c.shared_config_type_uuid, "type": c.type}

    def serialize_parameter_type(self, pt: ParameterType):
        result = {k: v for k, v in pt.__dict__.items() if v is not None}
        result["type"] = pt.type
        if pt.shared_object_type_uuid:
            result["shared_object_type_uuid"] = pt.shared_object_type_uuid
        return result



from nomnomdata.engine.components import Parameter, ParameterGroup, SharedConfig
from nomnomdata.engine.parameters import Boolean, Code, Enum, Int, Nested, String

DOSpace = SharedConfig(
    shared_config_type_uuid="DOSPC-NND3C",
    description="DigitalOcean Space configuration parameters.",
    alias="DigitalOcean Space",
    categories=["DigitalOcean", "Spaces"],
    parameter_groups=[
        ParameterGroup(
            Parameter(
                name="space",
                display_name="Space Name",
                description="Specify the name of the Space.",
                type=String(),
                required=True,
            ),
            Parameter(
                name="region",
                display_name="Region",
                description="Specify the region where the Space exists.",
                type=String(),
                required=True,
            ),
            Parameter(
                name="path_prefix",
                display_name="Path Prefix",
                description="Optional. Specify a path that will be prepended to the path specified in a Task to form the full path where files will be stored within the Space.",
                required=False,
                type=String(),
            ),
            Parameter(
                name="temp_path",
                display_name="Temp Path",
                description="Optional. Specify a full path within the Space where temporary files can be written.",
                required=False,
                type=String(),
            ),
            name="do_space_parameters",
            display_name="Space Parameters",
            description="",
        ),
    ],
)


FirebaseToDatabase = SharedConfig(
    shared_config_type_uuid="FB2DB-NND3C",
    description="Firebase collection to a relational database table configuration parameters.",
    alias="Firebase to Relational Database",
    categories=[
        "Firebase",
        "Loader",
        "Relational Database",
    ],
    parameter_groups=[
        ParameterGroup(
            Parameter(
                name="collection_name",
                display_name="Collection",
                description="Name of the collection in the Firebase app.",
                type=String(max=1024),
                required=True,
            ),
            Parameter(
                name="tracking_field",
                display_name="Tracking Field",
                description="Name of the field in the Collection to use for sorting and tracking documents processed.",
                required=False,
                type=String(),
            ),
            Parameter(
                name="tracking_field_type",
                display_name="Tracking Field Type",
                description="Select the data type of the Tracking Field.",
                required=False,
                type=Enum(choices=["NONE", "DATETIME", "STRING", "PUSHID"]),
                default="NONE",
            ),
            name="collection_parameters",
            display_name="Collection Parameters",
            description="Collection configuration information.",
        ),
        ParameterGroup(
            Parameter(
                name="load_pattern",
                display_name="Load Pattern",
                description="Select the pattern to use when loading the data.",
                type=Enum(choices=["INSERT"]),
                default="INSERT",
            ),
            Parameter(
                name="documentid_field",
                display_name="DocumentId Column",
                description="Map the DocumentId field to this column in the relational database.  Add a column with the same name to Column Parameters below.",
                type=String(),
                required=False,
            ),
            Parameter(
                name="date_processed_column",
                display_name="Date Processed Column",
                description="If specified, a date column representing when data was processed will be added to the relational database. Add a column with the same name to Column Parameters below.",
                type=String(),
                required=False,
            ),
            name="load_parameters",
            display_name="Load Parameters",
            description="Options for loading data into the relational database.",
        ),
        ParameterGroup(
            Parameter(
                name="column_parameters",
                display_name="Column Parameters",
                description="Details about each column within the relational database.",
                required=True,
                many=True,
                type=Nested(
                    Parameter(
                        name="column_name",
                        display_name="Column Name",
                        description="Specify the name of the column.",
                        type=String(max=128),
                        required=True,
                    ),
                    Parameter(
                        name="column_type",
                        display_name="Column Data Type",
                        description="Select the data type of the column. If type selected is not supported, the closest matching type will be used.",
                        type=Enum(
                            choices=[
                                "VARCHAR",
                                "INTEGER",
                                "BIGINT",
                                "DATETIME",
                                "DATE",
                                "TIME",
                                "TIMESTAMP",
                                "BOOLEAN",
                                "NUMERIC",
                                "FLOAT",
                                "BYTES",
                                "ARRAY",
                                "STRUCT",
                                "GEOGRAPHY",
                                # STRING
                            ]
                        ),
                        required=True,
                    ),
                    Parameter(
                        name="column_config",
                        display_name="Column Configuration",
                        description="Specify a configuration for the column. For example, VARCHAR could be (128), NUMERIC could be (12,2), ARRAY could be <STRING>, STRUCT could be <DATE, STRING>.",
                        type=String(),
                        required=False,
                    ),
                    Parameter(
                        name="json_path",
                        display_name="JSON Path",
                        description="Firebase field to map to the column. For example, ['store']['book']['title'] or store.book.title.",
                        type=String(),
                        required=False,
                    ),
                ),
            )
        ),
        ParameterGroup(
            Parameter(
                name="custom_parameter",
                display_name="Custom Parameters",
                description="Specify each parameter and value in quotes, separated by a colon.  Separate each pair with a comma and enclose all of the pairs in curly braces.",
                type=Code(),
                required=False,
            ),
            name="additional_parameters",
            display_name="Additional Parameters",
            description="Any additional parameters not described above.",
        ),
    ],
)

S3Bucket = SharedConfig(
    shared_config_type_uuid="S3BKT-NND3C",
    description="AWS S3 Bucket configuration parameters.",
    alias="S3 Bucket",
    categories=["Amazon", "AWS", "S3"],
    parameter_groups=[
        ParameterGroup(
            Parameter(
                name="bucket",
                display_name="Bucket Name",
                description="Specify the name of the S3 Bucket.",
                type=String(),
                required=True,
            ),
            Parameter(
                name="region",
                display_name="Region",
                description="Optional. Specify the region where the Bucket exists.",
                type=String(),
                required=False,
            ),
            Parameter(
                name="path_prefix",
                display_name="Path Prefix",
                description="Optional. Specify a path that will be prepended to the path specified in a Task to form the full path where files will be stored within the Bucket.",
                required=False,
                type=String(),
            ),
            Parameter(
                name="temp_path",
                display_name="Temp Path",
                description="Optional. Specify a full path within the Bucket where temporary files can be written.",
                required=False,
                type=String(),
            ),
            name="s3_bucket_parameters",
            display_name="Bucket Parameters",
            description="",
        ),
    ],
)

S3FileToDatabase = SharedConfig(
    shared_config_type_uuid="S3BTK-DB3MD",
    alias="S3 to Relational Database Load Options",
    description="Information needed when loading data from files stored on S3 into a relational database.",
    categories=[
        "S3",
        "Loader",
        "Relational Database",
    ],
    parameter_groups=[
        ParameterGroup(
            Parameter(
                name="bucket",
                display_name="Bucket",
                description="Specify the name of the S3 bucket that contains the files.",
                type=String(max=1024),
                required=True,
            ),
            Parameter(
                name="endpoint_url",
                display_name="Endpoint URL",
                description="For an S3 compatible bucket not hosted in AWS, specify the URL to reach it.",
                type=String(max=2048),
                required=False,
            ),
            Parameter(
                name="path",
                display_name="Path",
                description="Specify the path to the files within the S3 bucket.",
                type=String(max=2048),
                required=False,
            ),
            Parameter(
                name="s3_temp_space",
                display_name="Temporary File Storage Path",
                description="Specify the path within the S3 bucket where manifest files can be created.  If blank, nnd_tmp will be used.",
                type=String(max=2048),
                required=False,
            ),
            Parameter(
                name="region",
                display_name="Region",
                description="Optional. Specify the region where the Bucket exists.",
                type=String(),
                required=False,
            ),
            name="s3_parameters",
            display_name="S3 Parameters",
            description="Information about the S3 location where the files reside.",
        ),
        ParameterGroup(
            Parameter(
                name="load_pattern",
                display_name="Load Pattern",
                description="Select which pattern to use when loading the data from the files.",
                type=Enum(
                    choices=[
                        "INSERT",
                        "DELETE_ALL_INSERT_ALL",
                        "DELETE_ALL_INSERT_LAST",
                        "DELETE_INSERT",
                        "DELETE_INSERT_BATCH",
                        "DROP_AND_RECREATE",
                        "INSERT_IF_NOT_EXISTS",
                        "TRUNCATE_LOAD_ALL",
                        "TRUNCATE_LOAD_LAST",
                        "UPDATE_INSERT",
                        "UPDATE_INSERT_BATCH",
                    ]
                ),
                default="INSERT",
            ),
            Parameter(
                name="file_type",
                display_name="Data Format",
                description="Select the format of the data in the files.",
                type=Enum(choices=["JSON", "Delimited", "PARQUET"]),
                default="JSON",
            ),
            Parameter(
                name="compression",
                display_name="Compression Format",
                description="Select the compression, if any, applied to the files.",
                type=Enum(choices=["None", "gzip", "zip", "bzip2"]),
                default="None",
            ),
            Parameter(
                name="delimiter",
                display_name="Delimiter",
                description="Specify the character used to separate data values in files with delimited data format.",
                type=String(max=2),
                required=False,
            ),
            Parameter(
                name="null_value",
                display_name="Null Value",
                description="Specify the characters used to represent a null value in files with delimited data format.",
                type=String(max=8),
                required=False,
            ),
            Parameter(
                name="escape_character",
                display_name="Escape Character",
                description="Specify the character used to represent an escaped value in files with delimited data format.",
                type=String(max=2),
                required=False,
            ),
            Parameter(
                name="header_rows",
                display_name="Header Rows",
                description="Specify the number of lines to skip before the first row of data in files with delimited data format.",
                type=Int(),
                required=False,
            ),
            name="load_parameters",
            display_name="Load Parameters",
            description="Options describing how the data from the files should be loaded.",
        ),
        ParameterGroup(
            Parameter(
                name="partition_key",
                display_name="Partition Key",
                description="Optional. Specify the name of the single column whose values will be used to partition each file's data.",
                type=String(max=1024),
                required=False,
            ),
            Parameter(
                name="sort_key",
                display_name="Sort Keys",
                description="Optional. Specify one or more column names, with commas between each name, whose values will be used to sort each file's data.",
                type=String(max=2048),
                required=False,
            ),
            Parameter(
                name="primary_key",
                display_name="Primary Key",
                description="Optional. Specify the name of the single column whose values will be used to uniquely identify each row within each file's data.",
                type=String(max=2048),
                required=False,
            ),
            Parameter(
                name="merge_key",
                display_name="Merge Keys",
                description="Optional. Specify one or more column names, with commas between each name, whose combined values will be used to identify rows of data that will be merged together.",
                type=String(max=2048),
                required=False,
            ),
            Parameter(
                name="upsert_sort_key",
                display_name="Upsert Sort Keys",
                description="Optional. Specify one or more column names, with commas between each name, whose combined values will be used to set the order priority of rows of data that will be merged together.",
                type=String(max=2048),
                required=False,
            ),
            Parameter(
                name="merge_type",
                display_name="Merge Strategy",
                description="If column names are specified in the Merge Keys field, then rows with matching merge key values will be combined according to the strategy selected.",
                type=Enum(choices=["DISTINCT", "LOAD_ALL"]),
                default="DISTINCT",
                required=False,
            ),
            name="key_parameters",
            display_name="Key Parameters",
            description="Columns used to partition, sort, identify and merge data.",
        ),
        Parameter(
            name="column_parameters",
            display_name="Column Parameters",
            description="Details about each column within the relational database.",
            required=True,
            many=True,
            type=Nested(
                Parameter(
                    name="column_name",
                    display_name="Column Name",
                    description="Specify the name of the column.",
                    type=String(max=128),
                    required=True,
                ),
                Parameter(
                    name="column_type",
                    display_name="Column Data Type",
                    description="Select the data type of the column. If type selected is not supported, the closest matching type will be used.",
                    type=Enum(
                        choices=[
                            "VARCHAR",
                            "INTEGER",
                            "BIGINT",
                            "DATETIME",
                            "DATE",
                            "TIME",
                            "TIMESTAMP",
                            "BOOLEAN",
                            "NUMERIC",
                            "FLOAT",
                            "BYTES",
                            "ARRAY",
                            "STRUCT",
                            "GEOGRAPHY",
                            # STRING
                        ]
                    ),
                    required=True,
                ),
                Parameter(
                    name="column_config",
                    display_name="Column Configuration",
                    description="Specify a configuration for the column. For example, VARCHAR could be (128), NUMERIC could be (12,2), ARRAY could be <STRING>, STRUCT could be <DATE, STRING>.",
                    type=String(),
                    required=False,
                ),
                Parameter(
                    name="json_path",
                    display_name="Field Mapping",
                    description="For JSON formatted files, specify the path to the field that matches this column. For example, ['store']['book']['title'] or store.book.title.",
                    type=String(),
                    required=False,
                ),
            ),
        ),
        ParameterGroup(
            Parameter(
                name="custom_parameter",
                display_name="Custom Parameters",
                description="Specify each parameter and value in quotes, separated by a colon.  Separate each pair with a comma and enclose all of the pairs in curly braces.",
                type=Code(),
                required=False,
            ),
            name="additional_parameters",
            display_name="Additional Parameters",
            description="",
        ),
    ],
)


SQLToBigQueryParameters = SharedConfig(
    shared_config_type_uuid="SQLRW-BIGQRY",
    alias="SQL to BigQuery Load Options",
    description="Information needed when loading data from an SQL table into a BigQuery table.",
    categories=["Google", "BigQuery", "SQL"],
    parameter_groups=[
        ParameterGroup(
            Parameter(
                name="strategy",
                display_name="Load Pattern",
                description="Select which pattern to use when loading the data.",
                type=Enum(choices=["Append", "Overwrite", "Upsert"]),
                required=False,
                default="Append",
                help_header_id="Load Pattern",
            ),
            Parameter(
                name="upsert_column",
                display_name="Merge Keys",
                description="Optional. Specify one or more column names, with commas between each name, whose combined values will be used to identify rows of data that will be merged together.",
                type=String(),
                required=False,
                help_header_id="Merge Keys",
            ),
            Parameter(
                name="fields_mapping_autodetect",
                display_name="Autodetect Column Mappings",
                description="Automatically create the BigQuery table description.",
                type=Boolean(),
                required=False,
                default=True,
                help_header_id="Autodetect Column Mappings",
            ),
            Parameter(
                name="add_default_fields",
                display_name="Add Nom Nom Data Columns",
                description="Add the nnd_updated and nnd_created columns to the BigQuery table.",
                type=Boolean(),
                required=False,
                default=True,
                help_header_id="Add Nom Nom Data Columns",
            ),
            Parameter(
                name="column_parameters",
                display_name="Column Parameters",
                description="Details about each column within the tables.",
                required=True,
                many=True,
                type=Nested(
                    Parameter(
                        name="sql_row",
                        display_name="SQL Column Name",
                        description="Specify the name of the column.",
                        type=String(max=128),
                        required=True,
                    ),
                    Parameter(
                        name="bigquery_row",
                        display_name="BigQuery Column Name",
                        description="Specify the name of the column.",
                        type=String(max=128),
                        required=True,
                    ),
                    Parameter(
                        name="bigquery_row_type",
                        display_name="BigQuery Column Type",
                        description="Specify the type of the column.",
                        type=Enum(
                            choices=[
                                "STRING",
                                "BIGNUMERIC",
                                "BOOL",
                                "BYTES",
                                "DATE",
                                "DATETIME",
                                "FLOAT64",
                                "GEOGRAPHY",
                                "INT64",
                                "JSON",
                                "NUMERIC",
                                "STRING",
                                "TIME",
                                "TIMESTAMP",
                            ]
                        ),
                        required=True,
                    ),
                ),
            ),
        ),
    ],
)


VultrObjectStore = SharedConfig(
    shared_config_type_uuid="VULTR-OBJSR",
    description="Vultr Object Storage Bucket configuration parameters.",
    alias="Vultr Object Storage Bucket",
    categories=["Vultr", "Object Storage"],
    parameter_groups=[
        ParameterGroup(
            Parameter(
                name="bucket",
                display_name="Bucket Name",
                description="Specify the name of the Bucket.",
                type=String(),
                required=True,
            ),
            Parameter(
                name="region",
                display_name="Region",
                description="Specify the region where the Bucket exists.",
                type=Enum(choices=["ewr1"]),
                default="ewr1",
                required=True,
            ),
            Parameter(
                name="path_prefix",
                display_name="Path Prefix",
                description="Optional. Specify a path that will be prepended to the path specified in a Task to form the full path where files will be stored within the Bucket.",
                required=False,
                type=String(),
            ),
            Parameter(
                name="temp_path",
                display_name="Temp Path",
                description="Optional. Specify a full path within the Bucket where temporary files can be written.",
                required=False,
                type=String(),
            ),
            name="vultr_object_storage_parameters",
            display_name="Vultr Object Storage Parameters",
            description="",
        ),
    ],
)

KeyValue = SharedConfig(
    shared_config_type_uuid="GENRIC-KEYVAL",
    alias="Key and Value Pairs",
    description="One or more sets of key and value string pairs.",
    categories=[],
    parameter_groups=[
        Parameter(
            name="key_values",
            display_name="Pairs",
            description="",
            required=True,
            many=True,
            type=Nested(
                Parameter(
                    name="key",
                    display_name="Key",
                    description="Specify a key name.",
                    type=String(),
                    required=True,
                ),
                Parameter(
                    name="value",
                    display_name="Value",
                    description="Specify the value that corresponds to the key specified.",
                    type=String(),
                    required=True,
                ),
            ),
        ),
    ],
)



import logging
from abc import ABCMeta, abstractmethod
from typing import Any, Dict, List

from nomnomdata.engine.errors import LoadError, MissingParameters, ValidationError
from nomnomdata.engine.util import dict_union

py_type = type

logger = logging.getLogger(__name__)


class ParameterType(metaclass=ABCMeta):
    """
    ParameterType is the base class for all parameter types.
    Should never be directly instantiated.
    """

    @property
    @abstractmethod
    def type(self):
        pass

    @property
    def shared_object_type_uuid(self):
        return None

    def validate(self, val):
        return True

    def dump(self, val):
        return val

    def load(self, val):
        return val


class Rule:
    def __init__(self, name: str = "", value: List[str] = [""]):
        self.name = name
        self.value = value


class Parameter:
    """
    Parameter for use in a ParameterGroup, also contains options
    to control the look in the Nominode UI

    :param type:
        Set the type to allow validation of the parameter and
        so the UI understands how to render this parameter.
        Example valid parameter type :class:`~nomnomdata.engine.parameters.String`
    :param name:
        Parameter name, will be the key used in the final result passed to
        your function.
    :param display_name:
        The name of the parameter to be displayed to the user.
    :param help_header_id:
        Header ID in any MD file declared in upper scope that will be
        linked to this parameter.
    :param help_md_path:
        Full path to an MD file that will be linked
        to this parameter. Cannot be used with help_header_id.
    :param required:
        Setting this to true will require the user to set a value for this parameter,
        defaults to False
    :param description:
        The long form description the UI will diplay next
        to the parameter
    :param default:
        Default value of the parameter will be set as on task creation,
        valid values vary by the ParameterType you use, defaults to nothing
    """

    def __init__(
        self,
        type: ParameterType,
        name: str,
        display_name: str = "",
        help_header_id: str = None,
        help_md_path: str = None,
        required: bool = False,
        description: str = "",
        default: object = None,
        many: bool = False,
        categories: List[str] = None,
        show_on: Rule = None,
    ):
        if (
            not isinstance(type, ParameterType)
            and type is not None
            and issubclass(type, ParameterType)
        ):
            raise ValueError(f"{type} is not an instance, perhaps a forgotten ()?")
        elif not isinstance(type, ParameterType):
            raise ValueError(
                f"{type} is type {py_type(type)} when it must be an instance of a {ParameterType} subclass"
            )
        if name == "alias":
            raise ValueError(
                "alias is a protected parameter name, please use another name for your parameter"
            )
        self.type = type
        self.name = name
        if help_header_id and help_md_path:
            raise ValueError("Cannot use both help_header_id and help_md_path")
        if help_header_id:
            self.help = {"header_id": help_header_id}
        elif help_md_path:
            self.help = {"file": help_md_path}
        else:
            self.help = None
        self.display_name = display_name or self.name.capitalize()
        self.required = required
        self.description = description
        if default:
            type.validate(default)
        self.default = default
        self.many = many
        self.categories = [{"name": val} for val in categories] if categories else None
        self.show_on = show_on

    def validate(self, value: Any):
        self._verify_required(value)
        if self.many and value is not None:
            for v in value:
                self._validate(v)
        else:
            self._validate(value)
        return True

    def _validate(self, value: Dict[str, Any]):
        if value is not None:
            self.type.validate(value)
        return True

    def load(self, value: Any):
        if self.many:
            return [self._load(v) for v in value]
        else:
            return self._load(value)

    def _load(self, value: Any):
        try:
            result = self.type.load(value)
            return result
        except Exception as e:
            raise LoadError(f"Exception while loading parameter value: {value}") from e

    def _verify_required(self, val: Any):
        if self.required and val is None:
            raise MissingParameters()

    def __str__(self):
        return f"{self.type.__class__.__name__} Parameter"

    def __repr__(self):
        return f"<{self.type.__class__.__name__} Parameter>"


class ParameterGroup:
    """ParameterGroup acts as a logical and visual grouping for parameters

    :param name:
        Unique key for this group, defaults to "general"
    :warning: you cannot use parameter groups with the same name on one action
    :param display_name:
        Display name to render in the UI, defaults to "General Parameters"
    :param description: UI description for this parameter group, defaults to ""
    :param collapsed:
        If the UI should collapse this parameter group when rendering,
        defaults to False
    :param shared_parameter_group_uuid:
        UUID of the shared parameter group defined for this parameter,
        can be safely left undefined in 95% of cases
    """

    def __init__(
        self,
        *args: Parameter,
        name: str = "general",
        display_name: str = "General Parameters",
        description: str = "",
        collapsed: bool = False,
        shared_parameter_group_uuid: str = "",
    ):

        self.parameters = args
        self.collapsed = collapsed
        self.shared_parameter_group_uuid = shared_parameter_group_uuid
        self.description = description
        self.name = name
        self.display_name = display_name
        self.type = "group"


class NestedType(ParameterType):
    def __init__(self, parameter_groups: List[ParameterGroup]):
        self.parameter_groups = parameter_groups

    def load(self, value: Dict[str, Any]):
        union_params = dict_union(self.all_parameters, value)
        kwargs = {}
        for keyword, (param, val) in union_params.items():
            if keyword in ["alias", "connection_uuid", "shared_config_uuid"]:
                kwargs[keyword] = val
            elif not param:
                logger.warning(f"\tUnknown parameter '{keyword}', discarding")
            else:
                if param.many:
                    kwargs[keyword] = [param.type.load(v) for v in val] if val else []
                else:
                    kwargs[keyword] = param.type.load(val) if val is not None else val
        return kwargs

    def validate(self, value: Dict[str, Any]):
        if value:
            union_params = dict_union(self.all_parameters, value)
            for keyword, (param, val) in union_params.items():
                if keyword not in ["alias", "connection_uuid", "shared_config_uuid"]:
                    try:
                        param.validate(val)
                    except ValidationError as e:
                        e.add_key(keyword)
                        raise

    @property
    def all_parameters(self):
        results = {}
        for pg in self.parameter_groups:
            if isinstance(pg, Parameter):
                results[pg.name] = pg
            else:
                for p in pg.parameters:
                    results[p.name] = p
        return results


class Connection(NestedType):
    """
    Connections are generally used for authentication credentials,
    parameters will be stored encrypted on the Nominode instead of in plain text and can require
    seperate permissions to be viewed or used.


    :param connection_type_uuid:
        Unique ID associated with this Connection
    :param parameter_groups:
        ParameterGroups in this connection
    :param description:
        Description of the connection to be rendered on the UI, defaults to ""
    :param alias:
        Short(er) alias for the connection
    :param categories:
        Categories this connection belongs to, for easier sorting in the UI.
    """

    type = "connection"

    def __init__(
        self,
        connection_type_uuid: str,
        parameter_groups: List[ParameterGroup],
        description: str = "",
        alias: str = "",
        categories: List[str] = None,
    ):

        self.connection_type_uuid = connection_type_uuid
        self.description = description
        self.alias = alias
        self.categories = categories
        super().__init__(parameter_groups=parameter_groups)


class SharedConfig(NestedType):
    """
    Shared Config are generally used for authentication credentials,
    parameters will be stored encrypted on the Nominode instead of in plain text and can require
    seperate permissions to be viewed or used.


    :param shared_config_type_uuid:
        Unique ID associated with this Shared Config
    :param parameter_groups:
        ParameterGroups in this Shared Config
    :param description:
        Description of the Shared Config to be rendered on the UI, defaults to ""
    :param alias:
        Short(er) alias for the Shared Config
    :param categories:
        Categories this Shared Config belongs to, for easier sorting in the UI.
    """

    type = "shared_config"

    def __init__(
        self,
        shared_config_type_uuid: str,
        parameter_groups: List[ParameterGroup],
        description: str = "",
        alias: str = "",
        categories: List[str] = None,
    ):

        self.shared_config_type_uuid = shared_config_type_uuid
        self.description = description
        self.alias = alias
        self.categories = categories
        self.parameter_groups = parameter_groups



from wrapt import ObjectProxy


class LocalProxy(ObjectProxy):
    def __init__(self, name):
        super().__init__(None)

    def _set(self, proxy_obj):
        super().__init__(proxy_obj)


current_engine = LocalProxy("Engine Context")
current_nominode = LocalProxy("Nominode Context")
